from functools import partialmethod

base = 'http://lms.nthu.edu.tw'


class CourseRoute():

    rules = {
        'forum': 'forumlist',
        'homework': 'hwlist',
        'document': 'doclist',
        'forum_detail': 'forumlist',
        'homework_detail': 'hw',
        'document_detail': 'doc',
        'homework_handin_list_detail': 'hw_doclist',
    }

    rule_key = {
        'forum_detail': 'tid',
        'homework_detail': 'hw',
        'document_detail': 'cid',
        'homework_handin_list_detail': 'hw',
    }

    def __init__(self, course_id):
        self.base = '{}/course.php?courseID={}'.format(base, course_id)

    def gen_rule(self, func, uid=None):
        if uid:
            func += '_detail'
        path = '%s&f=%s' % (self.base, self.rules[func])
        if uid:
            path += '&{}={}'.format(self.rule_key[func], uid)
        return path

    homework = partialmethod(gen_rule, 'homework')
    homework_handin_list = partialmethod(gen_rule, 'homework_handin_list')
    document = partialmethod(gen_rule, 'document')
    forum = partialmethod(gen_rule, 'forum')


class Routes():

    rules = {
        'home': 'home.php',
        'profile': 'home/profile.php',
        'attach': 'sys/read_attach.php?id={attach_id}',
        'login_submit': 'sys/lib/ajax/login_submit.php',
        'post': 'sys/lib/ajax/post.php'
    }

    def __init__(self):
        for rule, value in self.rules.items():
            setattr(self, rule, '%s/%s' % (base, value))
        self.course = CourseRoute


route = Routes()



import os


_base_dir = os.path.expanduser('~')
_ilms_dir = os.path.join(_base_dir, '.ilms')
if not os.path.exists(_ilms_dir):
    os.makedirs(_ilms_dir)

_config_file = os.path.expanduser(os.path.join(_ilms_dir, 'ilms.json'))

__version__ = '0.0.3'



from ilms import parser
from ilms import exception
from ilms.route import route
from ilms.request import RequestProxyer
from ilms.utils import (
    unzip_all, check_is_download, stream_download, json_dump, json_load)

reqs = RequestProxyer()


class User:

    def __init__(self, user, pwd):
        self.user = user
        self.pwd = pwd
        self.email = None

    def check_login(self):
        try:
            resp = reqs.get(route.profile)
            parser.parse_profile(resp.text)
            return True
        except exception.PermissionDenied:
            return False

    def login(self):
        if self.check_login():
            return True
        resp = reqs.post(
                route.login_submit,
                data={'account': self.user, 'password': self.pwd})
        json = resp.json()
        if json['ret']['status'] == 'false':
            raise exception.LoginError
        self.email = json['ret']['email']
        reqs.save_session()
        return json


class Item():

    def __init__(self, raw, callee):
        self.raw = raw
        self.callee = callee
        self.uid = raw['id']
        self.insert_attrs(raw)

    def insert_attrs(self, attrs):
        for key, val in attrs.items():
            setattr(self, key, val)

    def download(self):
        for target in self.detail:
            download(target['id'])


class ItemContainer():

    def __init__(self, parse_result, instance, callee):
        self.items = [
            instance(item, callee=callee)
            for item in parse_result.result
        ]

    def __getitem__(self, x):
        return self.items[x]

    def __iter__(self):
        return self.items.__iter__()

    def get(self, x):
        return self.items[x]

    def find(self, **kwargs):
        for item in self.items:
            for k, v in kwargs.items():
                value = getattr(item, k)
                if value and isinstance(v, str) and v not in value:
                    break
            else:
                return item


class Handin(Item):

    @property
    def detail(self):
        if hasattr(self, '_detail'):
            return self._detail
        resp = reqs.get(
            route.course(self.callee.callee.id).document(self.uid))
        self._detail = parser.parse_homework_handin_detail(resp.text).result
        return self._detail

    def download(self, root_folder):
        folder_name = root_folder + '%s-%s' % (self.account_id, self.authour)
        if check_is_download(folder_name):
            return
        for target in self.detail:
            download(target['id'], folder=folder_name)
        unzip_all(folder_name)

    def __str__(self):
        return '<Homework Handin: %s>' % (self.authour)


class Homework(Item):

    @property
    def detail(self):
        if hasattr(self, '_detail'):
            return self._detail
        resp = reqs.get(
            route.course(self.callee.id).homework(self.uid))
        self._detail = parser.parse_homework_detail(resp.text).result
        return self._detail

    @property
    def handin_list(self):
        if hasattr(self, '_handin_list'):
            return self._handin_list
        resp = reqs.get(
            route.course(self.callee.id).homework_handin_list(self.uid))
        self._handin_list = ItemContainer(
            parser.parse_homework_handin_list(resp.text),
            instance=Handin,
            callee=self
        )
        return self._handin_list

    def download_handins(self):
        root_folder = './download/%s/' % self.title
        meta_path = root_folder + 'meta.json'
        done_lut = json_load(meta_path)
        try:
            for handin in self.handin_list:
                metadata = done_lut.get(handin.id)
                if (metadata
                   and metadata.get('date_string')
                   and handin.date_string > metadata['date_string']):
                    continue
                print(handin)
                handin.download(root_folder=root_folder)
                done_lut[handin.id] = {'last_update': handin.date_string}
        except Exception as e:
            print('Catch exception', e, 'while downloading', handin)
        finally:
            json_dump(done_lut, meta_path)

    def __str__(self):
        return '<Homework: %s>' % (self.title)


class Material(Item):

    @property
    def detail(self):
        if hasattr(self, '_detail'):
            return self._detail
        resp = reqs.get(
            route.course(self.callee.id).document(self.uid))
        self._detail = parser.parse_material_detail(resp.text).result
        return self._detail

    def __str__(self):
        return '<Material: %s, %s hits>' % (self.標題, self.人氣)


class Course(Item):

    def get_homeworks(self):
        resp = reqs.get(route.course(self.uid).homework())
        self.homeworks = ItemContainer(
            parser.parse_homework_list(resp.text),
            instance=Homework,
            callee=self
        )
        return self.homeworks

    def get_materials(self, download=False):
        resp = reqs.get(route.course(self.uid).document())
        self.materials = ItemContainer(
            parser.parse_material_list(resp.text),
            instance=Material,
            callee=self
        )
        return self.materials

    def get_forum_list(self, page=1):
        resp = reqs.get(
            route.course(self.uid).forum() + '&page=%d' % page)
        return parser.parse_forum_list(resp.text)

    def __str__(self):
        return '<Course: %s %s>' % (self.course_id, self.name.get('zh'))


class Core():

    def __init__(self, user):
        self.profile = None
        self.courses = None

    def get_profile(self):
        resp = reqs.get(route.profile)
        self.profile = parser.parse_profile(resp.text)
        return self.profile

    def get_courses(self):
        resp = reqs.get(route.home)
        self.courses = ItemContainer(
            parser.parse_course_list(resp.text),
            instance=Course,
            callee=self
        )
        return self.courses

    def get_post_detail(self, post_id):
        resp = reqs.post(route.post, data={'id': post_id})
        return parser.parse_post_detail(resp.json())


def download(attach_id, folder='download'):
    resp = reqs.get(route.attach.format(attach_id=attach_id), stream=True)
    stream_download(resp, folder)



import argparse

import ilms
from ilms.core import User, Core as iLms
from ilms.utils import get_account

parser = argparse.ArgumentParser()
parser.add_argument(
    'action', help=('login / download / view'))
parser.add_argument(
    '--courses', action='store_true', help='List all courses in this semester.')
parser.add_argument(
    '-c', '--course_id', type=str, help='Specific course to view.')
parser.add_argument(
    '-V', '--version', action='version', version=ilms.__version__)


def print_course_list(ilms):
    for cou in ilms.get_courses():
        print(cou)


def main(args=None):
    args = args or parser.parse_args()

    _account, _password = get_account()
    user = User(_account, _password)
    assert user.login()

    ilms = iLms(user)

    if args.courses:
        print_course_list(ilms)


# ilms download handin --course_id CS35700 --homework Homework1
# ilms download handin --course_id CS35700 --homework Homework1 --student 10502561 / 陳xx
# ilms download material --course_id CS35700 --num 5



import re
from datetime import datetime

# from html2text import html2text
from bs4 import BeautifulSoup

from ilms import exception


class ParseResult:

    def __init__(self, body=None):
        self.soup = BeautifulSoup(body, 'lxml') if body else None
        self.result = []
        self.extra = {}

pad = ['', ' ', '  ', '   ', ]


def need_login_check(f):
    def wrap(body, *args, **kwargs):
        if '權限不足' in body or 'No Permission!' in body:
            raise exception.PermissionDenied('尚未登入')
        return f(body, *args, **kwargs)
    return wrap


def parse_zh_en_course_name(course_name):
    course_name_en = re.findall('[A-Za-z()0-9 ]+', course_name)[-1]
    course_name_zh = course_name.replace(course_name_en, '')
    return {'en': course_name_en, 'zh': course_name_zh}


def fix_course_id_padding(course_id):
    dept = re.search('[A-Z]+', course_id).group()
    dept_padded = dept + pad[4 - len(dept)]
    return course_id.replace(dept, dept_padded)


def parse_datetime(date_string):
    return datetime.strptime(date_string, '%Y-%m-%d %H:%M:%S')


@need_login_check
def parse_profile(body):
    pr = ParseResult(body)
    name = pr.soup.select_one('#fmName').get('value')
    email = pr.soup.select_one('#fmEmail').get('value')
    pr.result = {'name': name, 'email': email}
    return pr


@need_login_check
def parse_course_list(body):
    pr = ParseResult(body)
    mnu = pr.soup.select('.mnu')[0]
    course_url_regex = re.compile('/course/(\d+)')

    for item in mnu.select('.mnuItem'):
        course_a = item.find('a')
        match = course_url_regex.match(course_a.get('href'))
        if not match:
            continue
        course_id = re.sub('[()]', '', item.find('span').text)
        pr.result.append({
            'id': match.group(1),
            'name': parse_zh_en_course_name(course_a.text),
            'course_id': course_id  # fix_course_id_padding(course_id)
        })
    return pr


@need_login_check
def parse_homework_list(body):
    pr = ParseResult(body)
    main = pr.soup.select_one('#main')
    if '目前尚無資料' in main.text:
        return pr

    for row in main.select('tr')[1:]:
        td = row.find_all('td')
        href = td[1].select_one('a').get('href')
        date = td[4].find('span').get('title')
        pr.result.append({
            'id': re.match('.*hw=(\d+).*', href).group(1),
            'title': td[1].text.strip(),
            'date_string': date,
            'date': parse_datetime(date)
        })
    return pr


@need_login_check
def parse_homework_detail(body):
    pr = ParseResult(body)
    tr = pr.soup.select('tr')

    def trs_helper(trs):
        for row in trs:
            k, v = row.select('td')
            yield k.text, v.text

    pr.result = {'title': pr.soup.select_one('#main span.curr').text.strip()}
    pr.result['extra'] = {
        k: v
        for i, (k, v) in enumerate(trs_helper(tr))
        if i not in [0, 5, 6, 7]
        # header, date, description, attachments
    }

    date = tr[5].select('td')[1].text + ':00'
    pr.result['date_string'] = date
    pr.result['date'] = parse_datetime(date)

    pr.result['content'] = tr[6].select('td')[1].text  # not rich text
    pr.result['links'] = [a.get('href') for a in tr[7].select('a')]

    td = tr[7].select('td')[1]
    attach_id_regex = re.compile('.*id=(\d+).*')
    pr.result['attachments'] = [
            {'name': a.text.strip(),
             'id': attach_id_regex.match(a.get('href')).group(1),
             'size': re.sub('[()]', '', span.text)}
            for a, span in zip(td.select('a'), td.select('span'))
        ]
    return pr


@need_login_check
def parse_homework_handin_list(body):
    pr = ParseResult(body)
    main = pr.soup.select_one('#main')
    if '目前尚無資料' in main.text:
        return pr

    # TODO: in not TA mode, some attrs will fail
    for row in main.select('tr')[1:]:
        td = row.select('td')
        href = td[1].select_one('a').get('href')
        date = td[5].find('span').get('title')
        status = {
            'status_id': td[6].find('span').get('id'),
            'text': td[6].text
        }
        score = {
            'score_id': td[7].select('.hidden div')[1].get('id'),
            'score_atag': td[7].select('.hidden div')[0].a
        }
        pr.result.append({
            'id': re.match('.*cid=(\d+).*', href).group(1),
            'title': td[1].text.strip(),
            'account_id': td[2].text,
            'authour': td[3].text,
            'status': status,
            'score': score,
            'date_string': date,
            'date': parse_datetime(date)
        })
    return pr


@need_login_check
def parse_homework_handin_detail(body):
    pr = ParseResult(body)
    main = pr.soup.select_one('#doc')
    if '目前尚無資料' in main.text:
        return pr

    attaches = main.select_one('.attach .block').select('div')
    for attach in attaches:
        a = attach.select('a')[1]
        hint = attach.select('.hint')[0]
        pr.result.append({
            'filename': a.get('title'),
            'id': a.get('href').split('=')[-1],
            'filesize': hint.text[1:-2]
        })

    return pr


@need_login_check
def parse_forum_list(body):
    pr = ParseResult(body)
    main = pr.soup.select_one('#main')
    if '目前尚無資料' in main.text:
        return pr

    for tr in main.select('tr')[1::2]:
        td = tr.select('td')
        pr.result.append({
            'id': td[0].text.strip(),
            'title': td[1].text.strip(),
            'count': td[2].find('span').text,
            'subtitle': td[3].text.strip()
        })

    page_info = pr.soup.select('.page span')
    if page_info:
        pages = len(page_info) - 2
        curr_page = pr.soup.select_one('.page .curr').text
        pr.extra = {
            'pages': pages if pages > 0 else 1,
            'curr_page': curr_page if pages > 0 else 1
        }

    return pr


@need_login_check
def parse_post_detail(json):
    pr = ParseResult()
    for item in json['posts']['items']:
        comment = {
            'name': item['name'],
            'date': item['date'],
            # 'note': html2text(item['note'])
            'note': item['note']
        }
        comment.update({
            'attachments': [(e['id'], e['srcName']) for e in item['attach']]
        }) if item['attach'] else None
        pr.result.append(comment)
    return pr


@need_login_check
def parse_material_list(body):
    pr = ParseResult(body)
    rows = pr.soup.select('tr')
    head = [e.text for e in rows[0].select('td')]
    for row in rows[1:]:
        item = {
            k: v.text.strip()
            for k, v in zip(head, row.select('td'))
        }
        item['id'] = row.select_one('a').get('href').split('=')[-1]
        pr.result.append(item)
    return pr


@need_login_check
def parse_material_detail(body):
    pr = ParseResult(body)
    title = pr.soup.select_one('.doc .title').text.strip()
    content = pr.soup.select_one('.doc .article').text.strip()

    attach = pr.soup.select_one('.attach .block')
    pr.extra = {'title': title, 'content': content}
    pr.result = [
        {
            'filename': a.get('title'),
            'id': a.get('href').split('=')[-1],
            'filesize': hint.text[1:-2]
        }
        for a, hint in zip(
            attach.select('a')[::2], attach.select('.hint'))
    ]
    return pr



from setuptools import setup

setup(
    name='ilms-nthu',
    packages=['ilms'],
    install_requires=[
        'requests',
        'beautifulsoup4',
        'lxml'
    ],
    entry_points={
        'console_scripts': ['ilms=ilms.cli:main'],
    },
    version='0.0.3a',
    description='iLms-NTHU API. An iLMS client for students, assistants and developers.',
    author='leVirve',
    author_email='gae.m.project@gmail.com',
    url='https://github.com/leVirve/iLms-nthu-API',
    license='MIT',
    platforms='any',
    keywords=['iLms', 'NTHU', 'API'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Customer Service',
        'Intended Audience :: System Administrators',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries',
        'Topic :: Text Processing',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],)



import os
import glob
import json
import pickle
import zipfile
import getpass
from pip._vendor.progress.bar import ShadyBar

import ilms


class ProgressBar(ShadyBar):

    suffix = ' %(percent).1f%%'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def calc_step(self, size):
        if size < 10:
            self.max = size
            return 1
        else:
            self.max = size // 10
            return 10


def get_account():
    if os.path.exists(ilms._config_file):
        _config = json.load(open(ilms._config_file))
        _account = _config.get('account')
        _password = _config.get('password')
        if _account and _password:
            return _account, _password

    _config = {'account': input('iLms account:'),
               'password': getpass.getpass(prompt='Password:')}
    _account = _config.get('account')
    _password = _config.get('password')
    with open(ilms._config_file, 'w') as f:
        f.write(json.dumps(_config, indent=4))

    return _account, _password


def unzip_all(folder_name):
    for zip_file in glob.glob('%s/*.zip' % folder_name):
        unzip(zip_file, folder_name)


def unzip(filepath, dest_folder):
    if filepath.endswith('.zip'):
        zip_ref = zipfile.ZipFile(filepath, 'r')
        zip_ref.extractall(dest_folder)
        zip_ref.close()


def check_is_download(folder):
    zip_type = ['*.zip', '*.rar']
    files = []
    for ztype in zip_type:
        files.extend(glob.glob('%s/%s' % (folder, ztype)))
    return files


def get_home_dir():
    _base_dir = os.path.expanduser('~')
    _ilms_dir = os.path.join(_base_dir, '.ilms')
    return _ilms_dir


def save_session(sess):
    base = ilms._ilms_dir
    with open(os.path.join(base, 'sess.pickle'), 'wb') as f:
        pickle.dump(sess, f, pickle.HIGHEST_PROTOCOL)


def load_session():
    base = ilms._ilms_dir
    filepath = os.path.join(base, 'sess.pickle')
    if not os.path.exists(filepath):
        return None
    with open(filepath, 'rb') as f:
        return pickle.load(f)


def json_load(filename):
    try:
        with open(filename) as f:
            return json.load(f)
    except:
        return {}


def json_dump(data, filename):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, 'w') as f:
        json.dump(data, f, indent=4)


def stream_download(stream_resp, folder='download'):
    filename = stream_resp.headers['content-disposition'].split("'")[-1]
    filesize = int(stream_resp.headers['content-length'])

    os.makedirs(folder, exist_ok=True)
    path = os.path.join(folder, filename)

    chunk_size = 1024
    progress = ProgressBar()
    progress.max = filesize // chunk_size
    with open(path, 'wb') as f:
        for chunk in stream_resp.iter_content(chunk_size=chunk_size):
            if chunk:
                f.write(chunk)
            progress.next()
    progress.finish()
    return filename



from functools import partialmethod

import requests

from ilms import utils


def encode_utf8(func):
    def f(*args, **kwargs):
        r = func(*args, **kwargs)
        r.encoding = 'utf8'
        return r
    return f


class RequestProxyer:

    def __init__(self, session=None):
        sess = utils.load_session()
        self.session = sess or session or requests.Session()

    def save_session(self):
        utils.save_session(self.session)

    @encode_utf8
    def request(*args, **kwagrs):
        return requests.get(*args, **kwagrs)

    @encode_utf8
    def request_auth(self, method, *args, **kwagrs):
        return self.session.request(method, *args, **kwagrs)

    get = partialmethod(request_auth, 'GET')
    post = partialmethod(request_auth, 'POST')
